package net.sf.mgp;

/*
 * Copyright 2008 MGP
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import java.io.File;
import java.net.URL;
import java.net.URLClassLoader;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.maven.artifact.Artifact;
import org.apache.maven.model.Resource;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.project.MavenProject;
import org.codehaus.plexus.util.StringUtils;
import org.codehaus.plexus.util.cli.CommandLineException;
import org.codehaus.plexus.util.cli.CommandLineUtils;
import org.codehaus.plexus.util.cli.Commandline;
import org.codehaus.plexus.util.cli.StreamConsumer;

/**
 * GWTCompiler goal
 * 
 * @goal compile
 * @phase process-sources
 * @requiresDependencyResolution compile
 * @description GWTCompiler goal
 * @author Maxim Gordienko
 */
public class CompileMojo extends AbstractMojo {
	private static final String GWTCOMPILER_CLASS_NAME = "com.google.gwt.dev.GWTCompiler";

	/**
	 * @parameter expression="${project}"
	 * @required
	 * @readonly
	 */
	private MavenProject project;

	/**
	 * Output directory for generated files
	 * 
	 * @parameter expression="${project.build.directory}/gwt"
	 * @required
	 */
	private File outputDirectory;

	/**
	 * Compiler JVM arguments
	 * 
	 * By default set headless mode and rise memory limit.
	 * 
	 * @see "http://code.google.com/p/google-web-toolkit-doc-1-5/wiki/FAQ_GWTCompilerX11Window"
	 * @see "http://code.google.com/p/google-web-toolkit-doc-1-5/wiki/FAQ_CompilerOutOfMemory"
	 * 
	 * @parameter default-value="-Djava.awt.headless=true -Xmx512M"
	 */
	private String jvmArgs;

	/**
	 * GWT project module (either single module or multiple modules required)
	 * 
	 * @see #modules
	 * @parameter
	 */
	private String module;

	/**
	 * GWT project modules (either single module or multiple modules required)
	 * 
	 * @see #module
	 * @parameter
	 */
	private String[] modules;

	/**
	 * JavaScript output style OBF[USCATED], PRETTY, or DETAILED
	 * 
	 * @parameter default-value="OBF"
	 */
	private String style;

	public void execute() throws MojoExecutionException {
		List<String> requestedModules = getModules();
		if (requestedModules.isEmpty()) {
			throw new MojoExecutionException("At least one module is required");
		}
		if (getLog().isDebugEnabled()) {
			getLog().debug("CompileMojo#execute()");
		}

		String classPath = getClassPath();
		for (String currentModule : requestedModules) {
			runCompiler(currentModule, classPath);
		}

		if (getLog().isDebugEnabled()) {
			getLog().debug("CompileMojo#execute() finished");
		}
	}

	private void runCompiler(String currentModule, String classPath) throws MojoExecutionException {
		Commandline cmd = new Commandline();
		cmd.setExecutable("java");
		if (!StringUtils.isEmpty(jvmArgs)) {
			cmd.addArguments(getArguments(jvmArgs));
		}
		cmd.createArg().setValue("-cp");
		cmd.createArg().setValue(classPath);

		cmd.createArg().setValue(GWTCOMPILER_CLASS_NAME);

		cmd.createArg().setValue("-out");
		cmd.createArg().setValue(outputDirectory.getAbsolutePath());
		cmd.createArg().setValue("-style");
		cmd.createArg().setValue(style);
		cmd.createArg().setValue(currentModule);

		if (getLog().isDebugEnabled()) {
			getLog().debug("Invoking compiler " + cmd.toString());
		}
		StreamConsumer logStreamConsumer = new StreamConsumer() {
			public void consumeLine(String line) {
				if (getLog().isDebugEnabled()) {
					getLog().debug("Compiler output: " + line);
				}
			}
		};

		try {
			int rc = CommandLineUtils.executeCommandLine(cmd, logStreamConsumer, logStreamConsumer);
			if (rc != 0) {
				throw new MojoExecutionException("GWT compiler exited with error: " + rc);
			}
		} catch (CommandLineException e) {
			throw new MojoExecutionException("Execution failed", e);
		}
	}

	private String getClassPath() {
		char sep = File.pathSeparatorChar;
		StringBuffer sb = new StringBuffer();

		for (String path : (List<String>) (project.getCompileSourceRoots())) {
			sb.append(path).append(sep);
		}

		for (Resource resource : (List<Resource>) (project.getResources())) {
			sb.append(resource.getDirectory()).append(sep);
		}

		for (String file : getCompileClasspathElements()) {
			sb.append(file).append(sep);
		}

		URLClassLoader currentClassLoader = (URLClassLoader) getClass().getClassLoader();
		URL[] originalUrls = currentClassLoader.getURLs();
		for (int i = 0; i < originalUrls.length; i++) {
			String url = originalUrls[i].toExternalForm();
			url = url.substring("file:".length());
			url = URLDecoder.decode(url);
			File file = new File(url);
			sb.append(file.getAbsolutePath()).append(sep);
		}

		return sb.toString();
	}

	/**
	 * @see MavenProject#getCompileClasspathElements()
	 */
	private List<String> getCompileClasspathElements() {
		List<String> list = new ArrayList<String>(project.getArtifacts().size());

		for (Artifact artifact : (Set<Artifact>) (project.getArtifacts())) {
			if (artifact.getArtifactHandler().isAddedToClasspath()) {
				if (Artifact.SCOPE_COMPILE.equals(artifact.getScope()) || Artifact.SCOPE_PROVIDED.equals(artifact.getScope())
						|| Artifact.SCOPE_SYSTEM.equals(artifact.getScope())) {

					File file = artifact.getFile();
					if (file != null) {
						list.add(file.getPath());
					} else if (getLog().isDebugEnabled()) {
						getLog().debug("No file for artifact " + artifact.getGroupId() + ":" + artifact.getArtifactId() + ":" + artifact.getVersion());
					}
				}
			}
		}
		return list;
	}

	private static final Pattern PARTS_PATTERN = Pattern.compile("\\S*\\\"(\\\\\"|[^\\\"])+\\\"\\S*|\\S+");

	String[] getArguments(String argString) {
		Matcher m = PARTS_PATTERN.matcher(argString);
		List<String> parts = new ArrayList<String>();
		int start = 0;
		while (m.find(start)) {
			parts.add(m.group());
			start = m.end();
		}
		return parts.toArray(new String[parts.size()]);
	}

	private List<String> getModules() {
		List<String> result = new ArrayList<String>();
		if (modules != null) {
			result.addAll(Arrays.asList(modules));
		}
		if (module != null) {
			result.add(module);
		}
		return result;
	}

}
